# Copyright 2017 Okera Inc. All Rights Reserved.
#
# Tests that should run on any configuration. The server auth can be specified
# as an environment variables before running this test.
# pylint: disable=bad-continuation,bad-indentation,global-statement,unused-argument
# pylint: disable=no-self-use
import time
import unittest
import json
import numpy

import os
import sys
sys.path.insert(
    0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

from okera.concurrency import OkeraWorkerException
from okera._thrift_api import TTypeId

from okera.tests import pycerebro_test_common as common
from okera.tests import create_test_data
import cerebro_common as cerebro
from instrumentation import measure_duration

retry_count = 0

class BasicTest(common.TestBase):
    def test_sparse_data(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            df = planner.scan_as_pandas("rs.sparsedata")
            self.assertEqual(96, len(df), msg=df)
            self.assertEqual(68, df['age'].count(), msg=df)
            self.assertEqual(10.0, df['age'].min(), msg=df)
            self.assertEqual(96.0, df['age'].max(), msg=df)
            self.assertEqual(b'sjc', df['defaultcity'].max(), msg=df)
            self.assertEqual(86, df['description'].count(), msg=df)

    def test_nulls(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            df = planner.scan_as_pandas("select string_col from rs.alltypes_null")
            self.assertEqual(1, len(df), msg=df)
            self.assertTrue(numpy.isnan(df['string_col'][0]), msg=df)

            df = planner.scan_as_pandas(
                "select length(string_col) as c from rs.alltypes_null")
            self.assertEqual(1, len(df), msg=df)
            self.assertTrue(numpy.isnan(df['c'][0]), msg=df)

    def test_timestamp_functions(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json("""
                select date_add('2009-01-01', 10) as c from okera_sample.sample""")
            self.assertTrue(len(json) == 2, msg=json)
            self.assertEqual('2009-01-11 00:00:00.000', str(json[0]['c']), msg=json)
            self.assertEqual('2009-01-11 00:00:00.000', str(json[1]['c']), msg=json)

    def test_duplicate_cols(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json("""
                select record, record from okera_sample.sample""")
            self.assertTrue(len(json) == 2, msg=json)
            self.assertEqual('This is a sample test file.', str(json[0]['record']),
                             msg=json)
            self.assertEqual('This is a sample test file.', str(json[0]['record_2']),
                             msg=json)

        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json("""
                select record, record as record_2, record from okera_sample.sample""")
            self.assertTrue(len(json) == 2, msg=json)
            self.assertEqual('This is a sample test file.', str(json[0]['record']),
                             msg=json)
            self.assertEqual('This is a sample test file.', str(json[0]['record_2']),
                             msg=json)
            self.assertEqual('This is a sample test file.', str(json[0]['record_2_2']),
                             msg=json)

    def test_large_decimals(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json("select num from rs.large_decimals2")
            self.assertTrue(len(json) == 6, msg=json)
            self.assertEqual('9012248907891233.020304050670',
                             str(json[0]['num']), msg=json)
            self.assertEqual('2343.999900000000', str(json[1]['num']), msg=json)
            self.assertEqual('900.000000000000', str(json[2]['num']), msg=json)
            self.assertEqual('32.440000000000', str(json[3]['num']), msg=json)
            self.assertEqual('54.230000000000', str(json[4]['num']), msg=json)
            self.assertEqual('4525.340000000000', str(json[5]['num']), msg=json)

        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            df = planner.scan_as_pandas("select num from rs.large_decimals2")
            self.assertTrue(len(df) == 6, msg=df)
            self.assertEqual('9012248907891233.020304050670',
                             str(df['num'][0]), msg=df)
            self.assertEqual('2343.999900000000', str(df['num'][1]), msg=df)
            self.assertEqual('900.000000000000', str(df['num'][2]), msg=df)
            self.assertEqual('32.440000000000', str(df['num'][3]), msg=df)
            self.assertEqual('54.230000000000', str(df['num'][4]), msg=df)
            self.assertEqual('4525.340000000000', str(df['num'][5]), msg=df)

    def test_date(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json("select * from datedb.date_csv")
            self.assertTrue(len(json) == 2, msg=json)
            self.assertEqual('Robert', str(json[0]['name']), msg=json)
            self.assertEqual(100, json[0]['id'], msg=json)
            self.assertEqual('1980-01-01', str(json[0]['dob']), msg=json)
            self.assertEqual('Michelle', str(json[1]['name']), msg=json)
            self.assertEqual(200, json[1]['id'], msg=json)
            self.assertEqual('1991-12-31', str(json[1]['dob']), msg=json)

        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            pd = planner.scan_as_pandas("select * from datedb.date_csv")
            self.assertTrue(len(pd) == 2, msg=pd)
            self.assertEqual(b'Robert', pd['name'][0], msg=pd)
            self.assertEqual(100, pd['id'][0], msg=pd)
            self.assertEqual('1980-01-01', str(pd['dob'][0]), msg=pd)
            self.assertEqual(b'Michelle', pd['name'][1], msg=pd)
            self.assertEqual(200, pd['id'][1], msg=pd)
            self.assertEqual('1991-12-31', str(pd['dob'][1]), msg=pd)

    def test_scan_as_json_max_records(self):
        sql = "select * from okera_sample.sample"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json(sql, max_records=1, max_client_process_count=1)
            self.assertTrue(len(json) == 1, msg='max_records not respected')
            json = planner.scan_as_json(sql, max_records=100, max_client_process_count=1)
            self.assertTrue(len(json) == 2, msg='max_records not respected')

    def test_scan_as_pandas_max_records(self):
        sql = "select * from okera_sample.sample"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            pd = planner.scan_as_pandas(sql, max_records=1, max_client_process_count=1)
            self.assertTrue(len(pd.index) == 1, msg='max_records not respected')
            pd = planner.scan_as_pandas(sql, max_records=100, max_client_process_count=1)
            self.assertTrue(len(pd.index) == 2, msg='max_records not respected')

    def test_scan_retry(self):
        global retry_count

        sql = "select * from okera_sample.sample"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            # First a sanity check
            pd = planner.scan_as_pandas(sql, max_records=1, max_client_process_count=1)
            self.assertTrue(len(pd.index) == 1, msg='test_scan_retry sanity check failed')

            # Patch scan_as_pandas to throw an IOError 2 times
            retry_count = 0
            def test_hook_retry(func_name, retries, attempt):
                if func_name != "plan":
                    return
                global retry_count
                retry_count = retry_count + 1
                if attempt < 2:
                    raise IOError('Fake Error')

            planner.test_hook_retry = test_hook_retry
            pd = planner.scan_as_pandas(sql, max_records=1, max_client_process_count=1)

            assert(retry_count == 3) # count = 2 failures + 1 success
            self.assertTrue(len(pd.index) == 1, msg='Failed to get data with retries')

    def test_worker_retry(self):
        global retry_count

        ctx = common.get_test_context()
        with common.get_worker(ctx) as worker:
            # First a sanity check
            v = worker.get_protocol_version()
            self.assertEqual('1.0', v)

            # Patch get_protocol_version to throw an IOError 2 times
            retry_count = 0
            def test_hook_retry(func_name, retries, attempt):
                if func_name != "get_protocol_version":
                    return
                global retry_count
                retry_count = retry_count + 1
                if attempt < 2:
                    raise IOError('Fake Error')

            worker.test_hook_retry = test_hook_retry
            v = worker.get_protocol_version()

            assert(retry_count == 3) # count = 2 failures + 1 success
            self.assertEqual('1.0', v)

    def test_overwrite_file(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            planner.execute_ddl("DROP TABLE IF EXISTS rs.dim")
            planner.execute_ddl("""CREATE EXTERNAL TABLE rs.dim
                (country_id INT, country_name STRING, country_code STRING)
                ROW FORMAT DELIMITED FIELDS TERMINATED BY ','
                LOCATION 's3://cerebro-datasets/starschema_demo/country_dim/'
                TBLPROPERTIES ('skip.header.line.count'='1')""")

            # Copy one version of the file into the target location
            cerebro.run_shell_cmd('aws s3 cp ' +\
                's3://cerebro-datasets/country_dim_src/country_DIM.csv ' +\
                's3://cerebro-datasets/starschema_demo/country_dim/country_DIM.csv')
            before = planner.scan_as_json('rs.dim')[0]
            self.assertEqual("France", before['country_name'], msg=str(before))

            # Copy another version. This file has the same length but a different
            # character. S3 maintains time in ms timestamp, so sleep a bit.
            time.sleep(1)

            cerebro.run_shell_cmd('aws s3 cp ' +\
                's3://cerebro-datasets/country_dim_src/country_DIM2.csv ' +\
                's3://cerebro-datasets/starschema_demo/country_dim/country_DIM.csv')
            i = 0
            while i < 10:
                after = planner.scan_as_json('rs.dim')[0]
                if 'france' in after['country_name']:
                    return
                self.assertEqual("France", after['country_name'], msg=str(after))
                time.sleep(.1)
                i = i + 1
            self.fail(msg="Did not updated result in time.")

    def test_scan_as_json_newline_delimiters(self):
        sql1 = '''select
         *
        from
        okera_sample.sample'''
        sql2 = '''select
        *
        from
        okera_sample.sample'''
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json(sql1, max_records=100, max_client_process_count=1)
            self.assertTrue(
                len(json) == 2,
                msg='could parse query with newline and space delimiters')
            json = planner.scan_as_json(sql2, max_records=100, max_client_process_count=1)
            self.assertTrue(
                len(json) == 2,
                msg='could parse query with newline delimiters')

    def test_scan_as_json_using_with_clause(self):
        sql1 = '''WITH male_customers AS
         (SELECT * FROM okera_sample.users WHERE gender = 'M')
         SELECT * FROM male_customers;'''
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json = planner.scan_as_json(sql1, max_records=100, max_client_process_count=1)
            self.assertTrue(
                len(json) == 100,
                msg='could parse query that starts with "with"')

    def test_scan_as_json_serialization(self):
        sql = "select * from rs.alltypes"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            json.loads(json.dumps(planner.scan_as_json(sql)))

    def test_das_6218(self):
        DB = "das_6218"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_db(conn, DB)
            self._create_all_types(conn, DB)

            # Verify schema from catalog, plan and data are all timestamp
            schema = conn.plan('%s.alltypes' % DB).schema
            self.assertEqual(schema.cols[10].type.type_id, TTypeId.TIMESTAMP_NANOS)
            df = conn.scan_as_pandas('SELECT timestamp_col FROM %s.alltypes' % DB)
            self.assertEqual(str(df.dtypes[0]), 'datetime64[ns, UTC]')
            catalog_schema = conn.list_datasets(DB, name='alltypes')[0].schema
            self.assertEqual(
                catalog_schema.cols[10].type.type_id, TTypeId.TIMESTAMP_NANOS)
            print(df)

            # Create a view that is proper with the explicit cast.
            conn.execute_ddl('''
                CREATE VIEW %s.v1(ts STRING)
                AS
                SELECT cast(timestamp_col AS STRING) FROM %s.alltypes''' % (DB, DB))
            # Verify schema from catalog, plan and data are all strings
            catalog_schema = conn.list_datasets(DB, name='v1')[0].schema
            self.assertEqual(catalog_schema.cols[0].type.type_id, TTypeId.STRING)
            schema = conn.plan('%s.v1' % DB).schema
            self.assertEqual(schema.cols[0].type.type_id, TTypeId.STRING)
            df = conn.scan_as_pandas('%s.v1' % DB)
            self.assertEqual(str(df.dtypes[0]), 'object')
            print(df)

            # We want to carefully construct a view that has mismatched types with
            # the view definition. The view just selects a timestamp columns but we
            # will force the catalog type to be string, instead of timestamp.
            # This forces the planner to produce an implicit cast.
            conn.execute_ddl('''
                CREATE EXTERNAL VIEW %s.v2(ts STRING)
                SKIP_ANALYSIS USING VIEW DATA AS
                "SELECT timestamp_col FROM %s.alltypes"''' % (DB, DB))
            # Verify schema from catalog, plan and data are all strings
            catalog_schema = conn.list_datasets(DB, name='v2')[0].schema
            self.assertEqual(catalog_schema.cols[0].type.type_id, TTypeId.STRING)
            schema = conn.plan('%s.v2' % DB).schema
            self.assertEqual(schema.cols[0].type.type_id, TTypeId.STRING)
            df = conn.scan_as_pandas('%s.v2' % DB)
            self.assertEqual(str(df.dtypes[0]), 'object')
            print(df)

    def test_zd_1633(self):
        DB = "zd_1633"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_db(conn, DB)
            conn.execute_ddl('''
                CREATE EXTERNAL TABLE %s.t(
                  s STRING)
                STORED AS TEXTFILE
                LOCATION 's3://cerebrodata-test/zd-1627/'
                ''' % DB)
            res = conn.scan_as_json('%s.t' % DB)
            # Default quote, only 1 row
            self.assertEqual(1, len(res))

            # No quote, should be 2 rows now
            conn.execute_ddl("ALTER TABLE %s.t SET SERDEPROPERTIES('quoteChar'='')" % DB)
            res = conn.scan_as_json('%s.t' % DB)
            self.assertEqual(2, len(res))

            # Recreate using table properties
            conn.execute_ddl('DROP TABLE %s.t' % DB)
            conn.execute_ddl('''
                CREATE EXTERNAL TABLE %s.t(
                  s STRING)
                STORED AS TEXTFILE
                LOCATION 's3://cerebrodata-test/zd-1627/'
                TBLPROPERTIES('okera.text-table.default-quote-char'='')
                ''' % DB)
            res = conn.scan_as_json('%s.t' % DB)
            self.assertEqual(2, len(res))

            # Explicit serde properties overrides table properties
            conn.execute_ddl('''
                ALTER TABLE %s.t SET SERDEPROPERTIES('quoteChar'='"')''' % DB)
            res = conn.scan_as_json('%s.t' % DB)
            self.assertEqual(1, len(res))

            # Table with two cols
            conn.execute_ddl('DROP TABLE %s.t' % DB)
            conn.execute_ddl('''
                CREATE EXTERNAL TABLE %s.t(
                  c1 STRING, c2 STRING)
                STORED AS TEXTFILE
                LOCATION 's3://cerebrodata-test/customers/c1/zd1633_2/'
                TBLPROPERTIES('skip.header.line.count'='1')
                ''' % DB)
            conn.execute_ddl(
                "ALTER TABLE %s.t SET SERDEPROPERTIES('field.delim'=',')" % DB)
            res = conn.scan_as_json('%s.t' % DB)
            self.assertEqual(1, len(res))

            # Remove quote handling
            conn.execute_ddl("ALTER TABLE %s.t SET SERDEPROPERTIES('quoteChar'='')" % DB)
            res = conn.scan_as_json('%s.t' % DB)
            self.assertEqual(2, len(res))
            self.assertEqual('"123', res[1]['c2'])

    def test_scan_delta_dummy_metadata(self):
        DB = "delta_dummy_scan_db"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_db(conn, DB)
            conn.execute_ddl('''
      CREATE EXTERNAL TABLE %s.airline_events_delta (
          col array<string>)
      WITH SERDEPROPERTIES (
          'path'='s3://cerebrodata-test/delta/airlines/airline_events/',
          'serialization.format'='1')
      STORED AS TEXTFILE
      LOCATION 's3://cerebrodata-test/delta/airlines/airline_events/'
      TBLPROPERTIES (
          'totalSize'='66153', 'numRows'='-1', 'rawDataSize'='-1',
          'okera.delta.infer-schema'='true',
          'COLUMN_STATS_ACCURATE'='false',
          'spark.sql.sources.schema.part.0'='{\"type\":\"struct\",\"fields\":[]}',
          'numFiles'='1', 'spark.sql.partitionProvider'='catalog',
          'spark.sql.sources.schema.numParts'='1', 'spark.sql.sources.provider'='delta',
          'spark.sql.create.version'='3.0.1')
                ''' % DB)
            res = conn.scan_as_json('%s.airline_events_delta' % DB)
            self.assertEqual(1326, len(res))

    def test_pandas_index(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as planner:
            # We scan a small table but with a tiny min task size, so that it is
            # guaranteed to generate more than one task. We then verify that our
            # overall row indices are correct.
            df = planner.scan_as_pandas("okera_sample.sample", min_task_size=1)

            indices = []
            for i, row in df.iterrows():
                indices.append(i)

            assert len(indices) == 2
            assert indices[0] == 0
            assert indices[1] == 1

    def test_zd1972(self):
        DB = "test_zd1972"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_db(conn, DB)
            conn.execute_ddl('''
            CREATE EXTERNAL TABLE %s.chase_kyc_event_v001 (
              `metadata` STRUCT<event_id:STRING,created_date:BIGINT>,
                `payload` STRUCT<party_key:STRING,customer_type:STRING,cdd:STRUCT<sowFromHighRiskIndustry:ARRAY<STRING>,countriesSignifWealthGen:ARRAY<STRING>,totalAnnualNetIncomeUsd:INT,estTotalNetWorth:INT,isSrcWealthCorroborated:BOOLEAN,srcWealthCorrobPercent:STRING>,spdd:STRUCT<pepHighRisk:BOOLEAN>,add:STRUCT<exceedingCashAggregate:BOOLEAN,exceedingEFTAggregate:BOOLEAN,additionalProdServices:ARRAY<STRING>>,exceptions:ARRAY<STRUCT<exceptionType:STRING>>,risk_profile:STRUCT<riskScore:STRING,risk_level:STRING>>,
                  `previouspayload` STRUCT<party_key:STRING,customer_type:STRING,cdd:STRUCT<sowFromHighRiskIndustry:ARRAY<STRING>,countriesSignifWealthGen:ARRAY<STRING>,totalAnnualNetIncomeUsd:INT,estTotalNetWorth:INT,isSrcWealthCorroborated:BOOLEAN,srcWealthCorrobPercent:STRING>,spdd:STRUCT<pepHighRisk:BOOLEAN>,add:STRUCT<exceedingCashAggregate:BOOLEAN,exceedingEFTAggregate:BOOLEAN,additionalProdServices:ARRAY<STRING>>,exceptions:ARRAY<STRUCT<exceptionType:STRING>>,risk_profile:STRUCT<riskScore:STRING,risk_level:STRING>>
                  )
            STORED AS AVRO
            LOCATION 's3a://cerebro-test-corey.sunwold/avro-testing/repro/dt=2021-02-25'
            TBLPROPERTIES ('avro.schema.url'='s3a://cerebro-test-corey.sunwold/avro/example-schema.json')
            ''' % DB)
            res = conn.scan_as_json('%s.chase_kyc_event_v001' % DB)
            self.assertEqual(1000, len(res))

            # Spot check a few enums
            self.assertEqual('MASS_MARKET', res[10]['previouspayload']['customer_type'])
            self.assertEqual('UNKNOWN', res[100]['previouspayload']['customer_type'])
            self.assertEqual('AFFLUENT', res[500]['previouspayload']['customer_type'])
            self.assertEqual('HIGH_WEALTH', res[900]['previouspayload']['customer_type'])

    def test_zd2001(self):
        DB = "test_zd2001"
        TBL = "bad_enums_1"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_db(conn, DB)
            ###
            # Avro Schema
            # {
            #    "type": "record",
            #    "name": "enums",
            #    "namespace": "enum_testing.avro",
            #    "fields":
            #    [
            #        {
            #            "name": "enum1",
            #            "type":
            #            {
            #                "type": "enum",
            #                "name": "test_enum",
            #                "symbols":
            #                [
            #                    "A",
            #                    "B",
            #                    "UNKNOWN"
            #                ]
            #            },
            #            "doc": "Name",
            #            "default": "UNKNOWN"
            #        }
            #    ]
            #}
            #
            #
            # bad_enums_1.avro contains a single row with a value of 10 which does not
            # map to any known enum value.
            ###
            conn.execute_ddl('''
            CREATE EXTERNAL TABLE %s.%s (
                enum1 STRING
            )
            STORED AS AVRO
            LOCATION "s3://cerebrodata-test/avro_test/enums/bad-enums/bad_enums_1.avro"
            ''' % (DB, TBL))
            with self.assertRaises(OkeraWorkerException) as ex_ctx:
                res = conn.scan_as_json('%s.%s' % (DB, TBL))

            self.assertTrue("Task failed due to an internal error." in str(ex_ctx.exception.error),
                msg=str(ex_ctx))

            # TODO The correct behavior is to deserialize the enum to the default value.
            # This is not implemented yet, and for now we will fail the query if we hit
            # a value we don't recognize. When this is fixed the below two assertions
            # should pass.
            #self.assertEqual(1, len(res))
            #self.assertEqual('UNKNOWN', res[0]['enum1'])

    def test_parquet_minmax_scanner_filter(self):
        db = "parquet_minmax_scanner_filter_db"
        tbl = "parquet_data"
        non_existent_id = 5802904565645645
        existing_id = 580389
        existing_small_int_id = 1
        non_existent_double_id = 2.5
        # we expect these queries to take less than this time
        max_duration = 40
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            create_test_data.create_parquet_table_with_stats(conn, db, tbl)
            # scan with a numeric predicate with the value that does not exist
            # so that we'd have to scan over the whole thing if the optimization
            # is not working.
            m = measure_duration('predicate with non_existent value duration')
            with m:
                res = conn.scan_as_json(
                'select trip_id from %s.%s where trip_id = %s' % (db, tbl, non_existent_id))
            print(m.duration)
            print(res)
            self.assertEqual(0, len(res))
            assert m.duration < max_duration

            m = measure_duration('predicate with existing value duration')
            with m:
                res = conn.scan_as_json(
                'select trip_id from %s.%s where trip_id = %s' % (db, tbl, existing_id))
            print(m.duration)
            print(res)
            self.assertEqual(1, len(res))
            assert m.duration < max_duration

            m = measure_duration('predicate with existing small int value duration')
            with m:
                res = conn.scan_as_json(
                'select trip_id from %s.%s where trip_id = %s' % (db, tbl, existing_small_int_id))
            print(m.duration)
            print(res)
            self.assertEqual(1, len(res))
            assert m.duration < max_duration

            m = measure_duration('predicate with non_existent double value duration')
            with m:
                res = conn.scan_as_json(
                'select trip_id from %s.%s where trip_id = %s' % (db, tbl, non_existent_double_id))
            print(m.duration)
            print(res)
            self.assertEqual(0, len(res))
            assert m.duration < max_duration

    # TODO: pushdown predicates with non-existing values do not have
    # consistent performance.
    @unittest.skip("Perf of predicate pushdown is unpredictable")
    def test_orc_minmax_scanner_filter(self):
        db = "orc_minmax_scanner_filter_db"
        tbl = "orc_data"
        non_existent_id = 5802904565645645
        existing_id = 580389
        existing_small_int_id = 1
        # we expect these queries to take less than 30 seconds
        max_duration = 30
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            create_test_data.create_orc_table_with_stats(conn, db, tbl)
            # scan with a numeric predicate with the value that does not exist
            # so that we'd have to scan over the whole thing if the optimization
            # is not working.

            m = measure_duration('predicate with non_existent value duration')
            with m:
                res = conn.scan_as_json(
                'select trip_id from %s.%s where trip_id = %s' % (db, tbl, non_existent_id))
            print(m.duration)
            print(res)
            self.assertEqual(0, len(res))

            assert m.duration < max_duration

            m = measure_duration('predicate with existing value duration')
            with m:
                res = conn.scan_as_json(
                'select trip_id from %s.%s where trip_id = %s' % (db, tbl, existing_id))
            print(m.duration)
            print(res)
            self.assertEqual(1, len(res))
            assert m.duration < max_duration

            m = measure_duration('predicate with existing small int value duration')
            with m:
                res = conn.scan_as_json(
                'select trip_id from %s.%s where trip_id = %s' % (db, tbl, existing_small_int_id))
            print(m.duration)
            print(res)
            self.assertEqual(1, len(res))
            assert m.duration < max_duration

    @unittest.skip("Need to fix bug. TDD")
    def test_zd1980(self):
        db = "zd_1980"
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._recreate_test_db(conn, db)

            conn.execute_ddl(
'''
CREATE EXTERNAL TABLE %s.chase_fraud_cards_screening_agent_decision_v001 (
  lifecycleid STRING COMMENT 'Unique identifier for a payment',
  decision STRING COMMENT 'The decision made by the agent. The value is configurable, it can be changed later. Actually the possible values are: manually_declined, auto_declined, fraud, manually_approved, auto_approved, genuine.',
  amount DECIMAL(15,2) COMMENT 'Amount, which is always greater than zero',
  currency STRING COMMENT 'Three-letter ISO currency codes eg GBP',
  country STRING COMMENT 'ISO country codes eg UK',
  username STRING COMMENT 'The user who made the decision',
  startedat BIGINT COMMENT 'The time that the decision was requested in UTC',
  finishedat BIGINT COMMENT 'The time that the decision was made in UTC'
)
PARTITIONED BY (dt DATE)
COMMENT 'The card transaction decision made by the agent through the system Feedzai Case Manager'
WITH SERDEPROPERTIES ('avro.schema.url'='s3a://cerebrodata-test/chase/chase.fraud.cards-screening-agent-decision-v001/dt=2021-03-12/chase.fraud.cards-screening-agent-decision-v001+0+0000000000.avro')
STORED AS AVRO
LOCATION 's3a://cerebrodata-test/chase/chase.fraud.cards-screening-agent-decision-v001'
''' % db)
            conn.execute_ddl('ALTER TABLE %s.chase_fraud_cards_screening_agent_decision_v001 RECOVER PARTITIONS' % db)

            conn.execute_ddl('''
CREATE EXTERNAL TABLE %s.event_1 (
  event_id STRING,
  event_type STRING,
  event_name STRING,
  entity_key STRING,
  partition_key STRING,
  checksum STRING,
  created_at BIGINT,
  card_transaction STRUCT<
    transaction_lifecycle_id:STRING,
    card_id:STRING,
    payment_device_details:STRUCT<
      device_id:STRING,
      funding_device_id:STRING,
      logical_card_id:STRING,
      device_type:STRING,
      token_details:STRUCT<
        token_number:STRING>>,
    party_key:STRING,
    state:STRING,
    card_messages:ARRAY<
      STRUCT<
        card_message_id:STRING,
        subscription_key:STRING,
        sequence_number:INT,
        matched_message_id:STRING,
        synthetic_trace_id:STRING,
        message_type:STRING,
        processing_category:STRING,
        pos_entry_mode:STRING,
        processing_code:STRING,
        acquirer_id:STRING,
        retrieval_reference_number:STRING,
        original_currency_amount:STRUCT<
          amount:DECIMAL(17,4),
          currency:STRING>,
        conversion_rate:DECIMAL(14,7),
        amount:STRUCT<
          amount:DECIMAL(17,4),
          currency:STRING>,
        content:STRUCT<
          message_type:STRING,
          primary_account_number:STRING,
          processing_code:STRING,
          processing_code_attributes:STRUCT<
            transaction_type_code:STRING,
            from_account_type_code:STRING,
            to_account_type_code:STRING>,
          system_trace_number:STRING,
          transaction:STRUCT<
            amounts:STRUCT<
              transaction:STRUCT<
                amount:DECIMAL(17,4),
                currency:STRING>,
              settlement:STRUCT<
                amount:DECIMAL(17,4),
                currency:STRING>,
              billing:STRUCT<
                amount:DECIMAL(17,4),
                currency:STRING>,
              fee:STRUCT<
                amount:DECIMAL(17,4),
                currency:STRING>,
              additional_amounts:ARRAY<
                STRUCT<
                  account_type:STRING,
                  amount_type:STRING,
                  transaction_sign:STRING,
                  amount:STRUCT<
                    amount:DECIMAL(17,4),
                    currency:STRING>>>,
            conversion_rate:DECIMAL(14,7)>,
            network_code:STRING,
            banknet_reference_number:STRING,
            dates:STRUCT<
              transaction_date:DATE,
              transaction_time:STRING,
              settlement_date:DATE,
              conversion_date:DATE,
              transmission_date_time:BIGINT>,
            retrieval_reference_number:STRING>,
          card:STRUCT<
            id:STRING,
            expiry_date:STRING>,
          acquirer:STRUCT<
            id:STRING,
            country_code:STRING>,
          merchant:STRUCT<
            name:STRING,
            address:STRUCT<
              city_name:STRING,
              state_or_country_code:STRING>,
            bankcard_phone:STRUCT<
              phone_number_dialed:STRING,
              abbreviation:STRING,
              call_duration:STRING,
              call_origin_city:STRING,
              call_origin_state_or_country_code:STRING>,
            terminal_id:STRING,
            category_code:STRING,
            acceptor_id_code:STRING>,
          pos:STRUCT<
            condition_code:STRING,
            additional_pos_detail:STRING,
            additional_pos_detail_attributes:STRUCT<
              pos_terminal_attendance:INT,
              pos_terminal_location:INT,
              pos_cardholder_presence:STRING,
              pos_card_presence:INT,
              pos_card_capture_capabilities:INT,
              pos_transaction_status:INT,
              pos_transaction_security:INT,
              cardholder_activated_terminal_level:INT,
              pos_card_data_terminal_input_capability_indicator:INT,
              pos_authorisation_life_cycle:INT,
              pos_country_code:STRING,
              pos_postal_code:STRING>,
            pos_entry_mode:STRING,
            pos_entry_mode_attributes:STRUCT<
              pan_entry_mode:STRING,
              pin_entry_mode:STRING>,
            extended_data_condition_codes:STRING>,
          additional_data:STRUCT<
            money_send_ref:STRING,
            multi_purpose_merchant_indicator:STRUCT<
              low_risk_merchant_indicator:STRING>,
            payment_initiation_channel:STRING,
            wallet_program_data:STRING,
            pan_mapping_file_information:STRING,
            trace_id:STRING,
            e_commerce_indicator:STRING,
            on_behalf_of_services:STRING,
            avs_response:STRING>,
          fraud_score_data:STRUCT<
            raw_data:STRING,
            way_4_risk_score:STRING,
            fraud_score:STRING,
            fraud_score_reason_code:STRING,
            rules_score:STRING,
            rules_reason_code_1:STRING,
            rules_reason_code_2:STRING>,
          wallet:STRUCT<
            account_number_indicator:STRING,
            account_number:STRING,
            expiry_date:STRING,
            token_requestor_id:STRING,
            wallet_program_data:STRING>,
          replacement_amounts:STRUCT<
            transaction:DECIMAL(17,4),
            settlement:DECIMAL(17,4),
            billing:DECIMAL(17,4)>,
          authentication:STRUCT<
            threeDS:STRUCT<
              authentication_protocol:STRING,
              directory_server_transaction_id:STRING>,
            ucaf:STRING,
            validationResults:STRUCT<
              cvc2_validation_result:STRING,
              ucaf_validation_result:STRING>,
            integratedCircuitCardData:STRING,
            pinValidation:STRING>,
          decline_reasons:ARRAY<
            STRUCT<
              reason:STRING,
              additional_data:MAP<STRING,STRING>>>,
          auth_code:STRING,
          auth_response_code:STRING,
          original_data_elements:STRUCT<
            message_type:STRING,
            system_trace_number:STRING,
            transmission_date_time:BIGINT,
            acquiring_institution_id:STRING,
            forwarding_institution_id:STRING>,
          paymentDevice:STRUCT<
            fundingDeviceId:STRING>,
          scheme_message_to_reverse_lifecycle_id:STRING,
          stipProcessor:STRING,
          stipIndicator:BOOLEAN,
          metadata:STRUCT<
            received_at:BIGINT>>,
        processing_result:STRUCT<
          response_code:STRING,
          auth_code:STRING,
          approved_amount:STRUCT<
            amount:DECIMAL(17,4),
            currency:STRING>,
          available_balance:STRUCT<
            amount:DECIMAL(17,4),
            currency:STRING>>,
        state:STRING,
        has_impacted_ledger:BOOLEAN,
        payment_device_details:STRUCT<
          device_id:STRING,
          funding_device_id:STRING,
          logical_card_id:STRING,
          device_type:STRING,
          token_details:STRUCT<
            token_number:STRING>>>>,
    ledger_transactions:ARRAY<
      STRUCT<
        ledger_transaction_id:STRING,
        subscription_key:STRING,
        ledger_type:STRING,
        original_currency_amount:STRUCT<
          amount:DECIMAL(17,4),
          currency:STRING>,
        average_conversion_rate:DECIMAL(16,9),
        amount:STRUCT<
          amount:DECIMAL(17,4),
          currency:STRING>,
        sign:STRING,
        state:STRING,
        transaction_code:STRING,
        value_date:BIGINT,
        booking_date:BIGINT,
        correlation_id:STRING>>,
    clearing_message:STRUCT<
      clearing_message_id:STRING,
      cleared_original_currency_amount:STRUCT<
        amount:DECIMAL(17,4),
        currency:STRING>,
        conversion_rate:DECIMAL(16,9),
        cleared_amount:STRUCT<
          amount:DECIMAL(17,4),
          currency:STRING>,
        acquirer_reference_data:STRING,
        content:STRUCT<
          id:STRING,
          transaction_date:STRING,
          retrieval_reference_number:STRING,
          merchant:STRUCT<
            name:STRING,
            address:STRUCT<
              city_name:STRING,
              state_code:STRING,
              country_code:STRING,
              post_code:STRING>,
            terminal_id:STRING,
            category_code:STRING,
            acceptor_id_code:STRING>,
          card:STRUCT<
            id:STRING,
            expiry_date:STRING,
            card_sequence_number:INT,
            primary_account_number:STRING>,
          amounts:STRUCT<
            transaction:STRUCT<
              amount:DECIMAL(17,4),
              currency:STRING>,
            settlement:STRUCT<
              amount:DECIMAL(17,4),
              currency:STRING>,
            billing:STRUCT<
              amount:DECIMAL(17,4),
              currency:STRING>,
            transaction_sign:STRING,
            interchange_fee:STRUCT<
              amount:DECIMAL(17,4),
              currency:STRING>,
            interchange_fee_sign:STRING,
            additional_amounts:ARRAY<
              STRUCT<
                amount_type:STRING,
                amount:STRUCT<
                  amount:DECIMAL(17,4),
                  currency:STRING>>>,
            settlement_conversion_rate:DECIMAL(16,9),
            billing_conversion_rate:DECIMAL(16,9)>,
          authentication_code:STRING,
          mapped_authorisation_processing_code:STRING,
          processing_code:STRING,
          pos:STRUCT<
            message_reason_code:STRING,
            pos_entry_mode:STRING>,
          function_code:INT,
          transaction_lifecycle_id:STRING,
          icc_data:STRING,
          ecommerce_security_level_indicator:STRING,
          mapping_service_account_number:STRING,
          wallet_identifier:STRING,
          data_record:STRING,
          message_number:INT,
          matching_indicator:BOOLEAN,
          file_id:STRING,
          acquirer_reference_data:STRING,
          created_at:BIGINT,
          file_key:STRING,
          row_number:INT,
          paymentDevice:STRUCT<
            fundingDeviceId:STRING>,
            wallet:STRUCT<
              mapping_service_account_number:STRING,
              wallet_identifier:STRING,
              token_requestor_id:STRING>>>,
        version:INT>)
    PARTITIONED BY (dt DATE)
    WITH SERDEPROPERTIES ('avro.schema.url'='s3a://cerebrodata-test/chase/chase.cards.card-transaction-event-v001/dt=2022-01-13/chase.cards.card-transaction-event-v001 34 0000008419.avro')
    STORED AS AVRO
    LOCATION 's3a://cerebrodata-test/chase/chase.cards.card-transaction-event-v001'
    ''' % db)
            conn.execute_ddl('ALTER TABLE %s.event_1 RECOVER PARTITIONS' % db)

            conn.execute_ddl('''
              CREATE VIEW %s.view_event_1 AS
              SELECT ** FROM %s.event_1''' % (db, db))

            conn.execute_ddl('''
CREATE VIEW %s.ren_chase_fraud_card_screening_agent_decision AS
SELECT
  dt seriesDate,
  lifecycleId paymentIdentifier,
  decision cardScreeningAgentDecisionCode,
  amount paymentAmount,
  currency currencyCode,
  country countryCode,
  username cardScreeningAgentDecisionUserName,
  startedAt cardScreeningAgentDecisionCodeStartTimestamp,
  finishedAt cardScreeningAgentDecisionCodeFinishTimestamp
FROM
  %s.chase_fraud_cards_screening_agent_decision_v001''' % (db, db))

            """
            Repro:
            [build]ubuntu@ip-10-1-11-17:~/okera/client/pycerebro/okera/test_scans$ rs-cat "select count(*) from zd_1980.event_1 where dt = cast('2021-01-01' as date)"
            rs-cat: Task failed due to an internal error.
            detail: Failed to parse file schema: Cannot save fixed schema. Schema length: 24687. Filename: s3a://cerebrodata-test/chase/chase.cards.card-transaction-event-v001/dt=2021-01-01/part-00000-886acf8c-2869-4ee0-b608-879d7f24f1bf-c000.avro
            """

if __name__ == "__main__":
    unittest.main()



# Copyright 2017 Okera Inc.
#

from __future__ import absolute_import

import builtins
import warnings
import logging
from logging import NullHandler
import string
import random

import six

from okera._thrift_api import (
    TExprType,
    TExprNodeType,
)

def get_logger_and_init_null(logger_name):
    logger = logging.getLogger(logger_name)
    logger.addHandler(NullHandler())
    return logger

log = get_logger_and_init_null(__name__)

def _random_id(prefix='', length=8):
    return prefix + ''.join(random.sample(string.ascii_uppercase, length))

def to_bytes(text):
    if isinstance(text, builtins.bytes):
        # We already have bytes, so do nothing
        return text
    if isinstance(text, list):
        # Convert a list of integers to bytes
        return builtins.bytes(text)
    # Convert UTF-8 text to bytes
    return builtins.bytes(str(text), encoding='utf-8')

def _escape(s):
    e = s
    e = e.replace('\\', '\\\\')
    e = e.replace('\n', '\\n')
    e = e.replace('\r', '\\r')
    e = e.replace("'", "\\'")
    e = e.replace('"', '\\"')
    log.debug('%s => %s', s, e)
    return e


def _py_to_sql_string(value):
    if value is None:
        return 'NULL'
    elif isinstance(value, six.string_types):
        return "'" + _escape(value) + "'"
    return str(value)

# Logging-related utils

def warn_deprecate(functionality='This', alternative=None):
    msg = ("{0} functionality in pyokera is now deprecated and will be removed "
           "in a future release".format(functionality))
    if alternative:
        msg += "; Please use {0} instead.".format(alternative)
    warnings.warn(msg, Warning)

class ExprUtils():
    """
    Utility class for handling expression ASTs
    """
    @staticmethod
    def convert_expr_ast_to_json(expr):
        """ Converts an expression AST to a python representation. This
            returns a dictionary that represents the AST. """
        root = None
        idx = 0
        while idx < len(expr.nodes):
            node, idx = ExprUtils.__convert_expr_ast_to_json_helper(expr.nodes, idx)
            if root is None:
                root = node
        return root

    @staticmethod
    def __convert_expr_ast_to_json_helper(nodes, idx):
        """ Utility for AST conversion. Thrift requires trees be serialized as
            lists, so this does the conversion."""
        node = nodes[idx]
        idx = idx + 1

        converted = {}
        converted['sql'] = node.sql
        converted['type'] = TExprType._VALUES_TO_NAMES[node.type]
        # we only want node_types for literals for now
        if converted['type'] == 'LITERAL':
            converted['node_type'] = TExprNodeType._VALUES_TO_NAMES[node.node_type]
        if node.num_children > 0:
            children = []
            for i in range(0, node.num_children):
                c, idx = ExprUtils.__convert_expr_ast_to_json_helper(nodes, idx)
                children.append(c)
            converted['children'] = children
        return converted, idx



# Copyright 2019 Okera Inc. All Rights Reserved.

import json
import pytest
import unittest

from okera._thrift_api import TAuthorizeQueryClient
from okera._thrift_api import TGetDatasetsParams
from okera._thrift_api import TRecordServiceException

from okera import policy_sync
from okera.tests import pycerebro_test_common as common
from okera.tests import snowflake_cte_test_cases

from sf_test import get_snowflake_test_context, get_snowflake

SKIP_LEVELS = ["smoke", "dev", "all", "checkin"]

@pytest.mark.skip(reason="Tests legacy, unmaintained version of policy sync")
class PolicySyncTests(common.TestBase):

    def test_coalescing_queue(self):
        q = policy_sync.CoalescingEventQueue(1, 5)
        q._enqueue('a', 1)
        q._enqueue('b', 0)
        # Not visible, delay of 1
        self.assertEqual(None, q._dequeue(0))
        # Only b visible
        self.assertEqual('b', q._dequeue(1))
        self.assertEqual(None, q._dequeue(1))
        # a visible @ t=2
        self.assertEqual('a', q._dequeue(2))
        self.assertEqual(None, q._dequeue(2))

        q._enqueue('a', 1)
        q._enqueue('a', 3)
        q._enqueue('a', 5)
        self.assertEqual(None, q._dequeue(1))
        self.assertEqual(None, q._dequeue(2))
        self.assertEqual(None, q._dequeue(3))
        self.assertEqual(None, q._dequeue(4))
        self.assertEqual(None, q._dequeue(5))
        self.assertEqual('a', q._dequeue(6))

    def _init_harness1(self, conn, db, tbl):
        grant1 = """
        GRANT SELECT ON TABLE %s.%s
        WHERE if (`int` > 100, true, false)
        TO ROLE sf_test_role1
        """ % (db, tbl)

        grant2 = """
        GRANT SELECT(`int`) ON TABLE %s.%s
        TO ROLE sf_test_role2
        """ % (db, tbl)

        grant3 = """
        GRANT SELECT ON TABLE %s.%s
        TRANSFORM snowflake_test.attr with tokenize()
        TO ROLE sf_test_role3
        """ % (db, tbl)

        self._recreate_test_role(conn, 'sf_test_role1', ['sf_testuser1'])
        self._recreate_test_role(conn, 'sf_test_role2', ['sf_testuser2'])
        self._recreate_test_role(conn, 'sf_test_role3', ['sf_testuser3'])
        conn.execute_ddl('DROP ATTRIBUTE IF EXISTS snowflake_test.attr')
        conn.execute_ddl('CREATE ATTRIBUTE snowflake_test.attr')
        conn.execute_ddl(
            'ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE `int` %s'\
                % (db, tbl, 'snowflake_test.attr'))
        conn.execute_ddl(
            'ALTER TABLE %s.%s ADD COLUMN ATTRIBUTE `string` %s'\
                % (db, tbl, 'snowflake_test.attr'))
        conn.execute_ddl(grant1)
        conn.execute_ddl(grant2)
        conn.execute_ddl(grant3)

    def snowflake_basic_test(self, db, tbl, sf_tbl):
        sf_db = sf_tbl.split('.')[0]
        sf_schema = '.'.join(sf_tbl.split('.')[0:2])

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            self._init_harness1(conn, db, tbl)
            users = ['sf_testuser1', 'sf_testuser2', 'sf_testuser3', 'noone']

            request = TGetDatasetsParams()
            request.dataset_names = ['{}.{}'.format(db, tbl)]
            request.with_schema = True
            okera_tbl = conn.service.client.GetDatasets(request).datasets[0]

            grants, pre_alters, row_filters, column_policies, post_alters, md5 = \
                policy_sync.generate_policies(
                    conn, users, db, tbl, okera_tbl, per_user_role_grants=True)
            policy_sync.debug_print(
                grants, pre_alters, row_filters, column_policies, post_alters)

            string_col_policy = None
            bool_col_policy = None
            if column_policies:
                for p in column_policies:
                    if p[0] == 'string':
                        string_col_policy = p[2]
                    if p[0] == 'boolean':
                        bool_col_policy = p[2]

            self.assertEqual('bb92f9b724f790e2bcf5d801cb942b7f', md5)
            expected_grants = [
                'CREATE ROLE IF NOT EXISTS "noone_ROLE"',
                'CREATE ROLE IF NOT EXISTS "okera_ROLE"',
                'CREATE ROLE IF NOT EXISTS "sf_testuser1_ROLE"',
                'CREATE ROLE IF NOT EXISTS "sf_testuser2_ROLE"',
                'CREATE ROLE IF NOT EXISTS "sf_testuser3_ROLE"',
                'REVOKE SELECT ON %s FROM ROLE "noone_ROLE"' % sf_tbl,
                'GRANT ROLE "noone_ROLE" TO USER "noone"',
                'GRANT SELECT ON %s TO ROLE "okera_ROLE"' % sf_tbl,
                'GRANT USAGE ON DATABASE %s TO ROLE "okera_ROLE"' % sf_db,
                'GRANT USAGE ON SCHEMA %s TO ROLE "okera_ROLE"' % sf_schema,
                'GRANT SELECT ON %s TO ROLE "sf_testuser1_ROLE"' % sf_tbl,
                'GRANT USAGE ON DATABASE %s TO ROLE "sf_testuser1_ROLE"' % sf_db,
                'GRANT USAGE ON SCHEMA %s TO ROLE "sf_testuser1_ROLE"' % sf_schema,
                'GRANT ROLE "sf_testuser1_ROLE" TO USER "sf_testuser1"',
                'GRANT SELECT ON %s TO ROLE "sf_testuser2_ROLE"' % sf_tbl,
                'GRANT USAGE ON DATABASE %s TO ROLE "sf_testuser2_ROLE"' % sf_db,
                'GRANT USAGE ON SCHEMA %s TO ROLE "sf_testuser2_ROLE"' % sf_schema,
                'GRANT ROLE "sf_testuser2_ROLE" TO USER "sf_testuser2"',
                'GRANT SELECT ON %s TO ROLE "sf_testuser3_ROLE"' % sf_tbl,
                'GRANT USAGE ON DATABASE %s TO ROLE "sf_testuser3_ROLE"' % sf_db,
                'GRANT USAGE ON SCHEMA %s TO ROLE "sf_testuser3_ROLE"' % sf_schema,
                'GRANT ROLE "sf_testuser3_ROLE" TO USER "sf_testuser3"',
            ]

            expected_row_filters = '''
CREATE OR REPLACE ROW ACCESS POLICY %s_row_policy AS (int BIGINT) RETURNS boolean ->
CASE
    WHEN current_user() in ('okera', 'sf_testuser2', 'sf_testuser3') THEN TRUE
    WHEN current_user() in ('sf_testuser1') THEN iff("INT" > 100, TRUE, FALSE)
    ELSE FALSE
END''' % sf_tbl

            expected_string_col_policy = '''
CREATE OR REPLACE MASKING POLICY %s_string_masking_policy AS (string STRING) RETURNS STRING ->
CASE
    WHEN current_user() in ('okera', 'sf_testuser1') THEN string
    WHEN current_user() in ('sf_testuser2') THEN NULL
    WHEN current_user() in ('sf_testuser3') THEN okera_udfs.public.tokenize("STRING", last_query_id())
    ELSE NULL
END''' % sf_tbl

            expected_boolean_col_policy = '''
CREATE OR REPLACE MASKING POLICY %s_boolean_masking_policy AS (boolean BOOLEAN) RETURNS BOOLEAN ->
CASE
    WHEN current_user() in ('okera', 'sf_testuser1', 'sf_testuser3') THEN boolean
    WHEN current_user() in ('sf_testuser2') THEN NULL
    ELSE NULL
END''' % sf_tbl

            expected_pre_alters = [
                'ALTER TABLE %s DROP ROW ACCESS POLICY %s_row_policy' % (sf_tbl, sf_tbl),
                'ALTER TABLE %s MODIFY COLUMN VARCHAR UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_VARCHAR_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN STRING UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_STRING_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN TEXT UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_TEXT_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN SMALLINT UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_SMALLINT_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN INT UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_INT_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN BIGINT UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_BIGINT_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN INTEGER UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_INTEGER_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN DOUBLE UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_DOUBLE_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN NUMERIC UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_NUMERIC_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN NUMBER UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_NUMBER_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN DECIMAL UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_DECIMAL_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN TIMESTAMP UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_TIMESTAMP_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN CHAR UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_CHAR_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN BOOLEAN UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_BOOLEAN_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN BINARY UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_BINARY_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN VARBINARY UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_VARBINARY_masking_policy' % sf_tbl,
                'ALTER TABLE %s MODIFY COLUMN REAL UNSET MASKING POLICY' % sf_tbl,
                'DROP MASKING POLICY %s_REAL_masking_policy' % sf_tbl]

            expected_post_alters = [
                'ALTER TABLE %s ADD ROW ACCESS POLICY %s_row_policy ON(int)' % (sf_tbl, sf_tbl),
                'ALTER TABLE %s MODIFY COLUMN bigint SET MASKING POLICY %s_bigint_masking_policy' % (sf_tbl, sf_tbl)]

            self.assertEqual(expected_grants, grants)
            self.assertEqual(expected_pre_alters, pre_alters[0:len(expected_pre_alters)])
            self.assertEqual(expected_row_filters, row_filters)
            self.assertEqual(expected_string_col_policy, string_col_policy)
            self.assertEqual(expected_boolean_col_policy, bool_col_policy)
            self.assertEqual(
                expected_post_alters, post_alters[0:len(expected_post_alters)])

    def test_snowflake_policy_gen(self):
        self.maxDiff = None
        self.snowflake_basic_test(
            'jdbc_test_snowflake', 'all_types', 'DEMO_DB.JDBC_TEST.ALL_TYPES')

    def test_snowflake_test_get_users(self):
        self.maxDiff = None
        sf_ctx = get_snowflake_test_context(username='SFV3ADMIN')
        with get_snowflake(sf_ctx) as conn:
            users = policy_sync.get_snowflake_users(conn, before_role='ACCOUNTADMIN')
            self.assertTrue('OKERA_ANALYST' in users)

    def test_compute_users(self):
        users = ['a', 'b', 'c']
        props = {}

        self.assertEqual(['a', 'b', 'c'], policy_sync.get_sync_users(users, None))
        self.assertEqual(['a', 'b', 'c'], policy_sync.get_sync_users(users, props))

        props[policy_sync.TBL_PROP_USERS_LIST] = 'd,e'
        self.assertEqual(['d', 'e'], policy_sync.get_sync_users(users, props))

        props = {}
        props[policy_sync.TBL_PROP_USERS_ADDITIONAL] = 'd,e'
        self.assertEqual(['a', 'b', 'c', 'd', 'e'], policy_sync.get_sync_users(
            users, props))

        props = {}
        props[policy_sync.TBL_PROP_USERS_BLACKLIST] = 'd,e'
        self.assertEqual(['a', 'b', 'c'], policy_sync.get_sync_users(
            users, props))
        props[policy_sync.TBL_PROP_USERS_BLACKLIST] = 'b,d,e'
        self.assertEqual(['a', 'c'], policy_sync.get_sync_users(
            users, props))

        props = {}
        props[policy_sync.TBL_PROP_USERS_WHITELIST] = 'd,e'
        self.assertEqual([], policy_sync.get_sync_users(
            users, props))
        props[policy_sync.TBL_PROP_USERS_WHITELIST] = 'b,d,e'
        self.assertEqual(['b'], policy_sync.get_sync_users(
            users, props))

        props = {}
        props[policy_sync.TBL_PROP_USERS_ADDITIONAL] = 'd,e'
        self.assertEqual(['a', 'b', 'c', 'd', 'e'], policy_sync.get_sync_users(
            users, props))
        props[policy_sync.TBL_PROP_USERS_WHITELIST] = 'a,d,e'
        self.assertEqual(['a', 'd', 'e'], policy_sync.get_sync_users(
            users, props))
        props[policy_sync.TBL_PROP_USERS_BLACKLIST] = 'e'
        self.assertEqual(['a', 'd'], policy_sync.get_sync_users(
            users, props))

    def test_snowflake_policy_e2e(self):
        self.maxDiff = None
        sf_ctx = get_snowflake_test_context(username='SFV3ADMIN')
        ctx = common.get_test_context()

        DB = 'snowflake_policy_sync_test'
        TBL = 'all_types'
        CXN = 'snowflake_policy_sync_test_cxn'

        drc = self._get_snowflake_data_reg_connection_obj(CXN)
        drc.default_catalog = "POLICY_SYNC_TEST"
        drc.default_schema = "PUBLIC"
        drc.connection_properties = {'defaultDb':'PUBLIC'}

        with common.get_planner(ctx) as conn:
            conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % DB)
            conn.execute_ddl('DROP DATACONNECTION IF EXISTS %s' % CXN)
            drcs = conn.manage_data_reg_connection("CREATE", drc)
            conn.execute_ddl("""CREATE DATABASE %s DBPROPERTIES(
                'okera.connection.name' = '%s',
                'okera.autotagger.skip'='true')""" % (DB, CXN))
            conn.execute_ddl('ALTER DATABASE %s LOAD DEFINITIONS()' % DB)
            self._init_harness1(conn, DB, TBL)

            with get_snowflake(sf_ctx) as sf_conn:
                policy_sync.sync_policies(
                    conn, sf_conn, DB, TBL, 'ACCOUNTADMIN',
                    role_pattern='"OKERA_CI_TEST_%s_ROLE"')

    def _init_adhoc(self, conn, auto_sync):
        DB = 'snowflake_policy_sync_adhoc'
        TBL = 'patient'
        CXN = 'snowflake_policy_sync_adhoc_test_cxn'

        drc = self._get_snowflake_data_reg_connection_obj(CXN)
        drc.default_catalog = "POLICY_SYNC_ADHOC"
        drc.default_schema = "HEALTHCARE"
        drc.connection_properties = {'defaultDb':'PUBLIC'}

        self._recreate_test_role(
            conn, 'sf_test_adhoc_role',
            ['sf_testuser1', 'SFV3TEST'])

        conn.execute_ddl('DROP DATABASE IF EXISTS %s CASCADE' % DB)
        conn.execute_ddl('DROP DATACONNECTION IF EXISTS %s' % CXN)
        drcs = conn.manage_data_reg_connection("CREATE", drc)
        conn.execute_ddl("""CREATE DATABASE %s DBPROPERTIES(
            'okera.connection.name' = '%s',
            'okera.policy-sync.enabled' = 'true',
            'okera.policy-sync.auto-sync' = '%s',
            'okera.autotagger.skip'='true')""" % (DB, CXN, auto_sync))
        conn.execute_ddl('ALTER DATABASE %s LOAD DEFINITIONS()' % DB)

    def snowflake_sync_adhoc(self, auto_sync):
        ctx = common.get_test_context()

        self.maxDiff = None
        sf_ctx = get_snowflake_test_context(username='SFV3TEST')
        sf_query = 'SELECT * FROM POLICY_SYNC_ADHOC.HEALTHCARE.PATIENT LIMIT 2'

        with get_snowflake(sf_ctx) as sf_conn:
            with common.get_planner(ctx) as conn:
                self._init_adhoc(conn, auto_sync)

                # Basic just loading it, no roles, policies or attributes
                if not auto_sync:
                    conn.execute_ddl('ALTER TABLE snowflake_policy_sync_adhoc.patient EXECUTE policy_sync')
                md5_0 = conn.list_datasets(db=DB, name=TBL)[0].metadata['policy-sync.md5']
                if not auto_sync:
                    conn.execute_ddl('ALTER TABLE snowflake_policy_sync_adhoc.patient EXECUTE policy_sync')
                md5_1 = conn.list_datasets(db=DB, name=TBL)[0].metadata['policy-sync.md5']
                self.assertEqual(md5_0, md5_1)

                # Initialize this connection to use the OKERA role created by our policy
                # sync
                sf_conn.query_request('USE ROLE %s' % '"OKERA_ADHOC_SFV3TEST_ROLE"')

                # Should fail, user has no access
                with self.assertRaises(Exception) as ex_ctx:
                    sf_conn.query_request(sf_query)
                self.assertTrue(
                    'does not exist or not authorized' in str(ex_ctx.exception))

                # Grant full table access
                conn.execute_ddl("""
                    GRANT SELECT ON TABLE %s.%s
                    TO ROLE sf_test_adhoc_role
                """ % (DB, TBL))
                if not auto_sync:
                    conn.execute_ddl('ALTER TABLE snowflake_policy_sync_adhoc.patient EXECUTE policy_sync')
                md5_2 = conn.list_datasets(db=DB, name=TBL)[0].metadata['policy-sync.md5']
                self.assertTrue(md5_0 != md5_2)
                result = sf_conn.query_request(sf_query)
                self.assertEqual(result['rowset'][0][2], 'Cristy')
                self.assertEqual(result['rowset'][1][4], '337-35-7356')

                # Revoke and try again
                conn.execute_ddl("""
                    REVOKE SELECT ON TABLE %s.%s
                    FROM ROLE sf_test_adhoc_role
                """ % (DB, TBL))
                if not auto_sync:
                    conn.execute_ddl('ALTER TABLE snowflake_policy_sync_adhoc.patient EXECUTE policy_sync')
                with self.assertRaises(Exception) as ex_ctx:
                    sf_conn.query_request(sf_query)
                self.assertTrue('does not exist or not authorized' in str(ex_ctx.exception))

    @unittest.skip("Adhoc example. Cannot be run concurrently.")
    def test_snowflake_sync_adhoc(self):
        self.maxDiff = None
        self.snowflake_sync_adhoc(False)

    @unittest.skip("Adhoc example. Cannot be run concurrently.")
    def test_snowflake_auto_sync_adhoc(self):
        self.maxDiff = None
        self.snowflake_sync_adhoc(True)

    def test_catalog_sync(self):
        # Test only makes sense to run if dry_run=True
        self.maxDiff = None
        ctx = common.get_test_context()
        sf_ctx = get_snowflake_test_context(username='SFV3ADMIN')
        with get_snowflake(sf_ctx) as sf_conn:
            with common.get_planner(ctx) as conn:
                self._init_adhoc(conn, False)
                policy_sync.sync_catalog_loop(
                    conn, sf_conn, 'ACCOUNTADMIN', min_sync_time_secs=15, iters=1,
                    dry_run=True)

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import os
import sys

import configargparse

from okera import context
from okera import policy_sync
from okera.tests import pycerebro_test_common as common
from okera._thrift_api import TAuthorizeQueryClient, TGetDatasetsParams

ROOT_TOKEN = os.environ['OKERA_HOME'] + '/integration/tokens/cerebro.token'
NIGHTLY_PLANNER = 'nightly.internal.okera.rocks'

parser = configargparse.ArgParser()
parser.add("--table", dest="table", default=None,
            help="[db].[table] to generate policy sync for")
parser.add("--users", dest="users", default=None,
            help="Comma separated list of users to generate policy sync for.")

parser.add("--single-role", dest="single_role", default=None,
            help="Name of the role that should be used for all users.")
parser.add("--per-user-role", dest="per_user_role_grants", default=False,
            help="Grant to a per user role.")

parser.add("--nightly", dest="nightly", default=False,
            help="If true, connect to nightly (vs local).")
options = parser.parse_args()

def error(msg):
  print("error: " + msg)
  sys.exit(1)

def get_conn():
    if bool(options.nightly):
        ctx = context()
        ctx.enable_token_auth(token_file=ROOT_TOKEN)
        return ctx.connect(NIGHTLY_PLANNER)
    else:
        ctx = common.get_test_context()
        return common.get_planner(ctx)

if __name__ == "__main__":
    if options.table is None:
        error("Must specify table.")
    if len(options.table.split('.')) != 2:
        error("Table must be [db].[table]")
    if options.users is None:
        error("Must specify users.")

    users = []
    for u in options.users.split(','):
        users.append(u.strip())
    db = options.table.split('.')[0]
    tbl = options.table.split('.')[1]

    with get_conn() as conn:
        request = TGetDatasetsParams()
        request.dataset_names = ['{}.{}'.format(db_name, tbl_name)]
        request.with_schema = True
        okera_tbl = planner_conn.service.client.GetDatasets(request).datasets[0]
        grants, pre_alters, row_filters, column_policies, post_alters, _ = \
            policy_sync.generate_policies(
            conn, users, db, tbl, okera_tbl,
            single_grant_role=options.single_role,
            per_user_role_grants=bool(options.per_user_role_grants))
        policy_sync.debug_print(grants, pre_alters, row_filters, column_policies, post_alters)



# Copyright 2021 Okera Inc. All Rights Reserved.
#
# Latency perf suite.
#
# pylint: disable=global-statement
# pylint: disable=no-self-use

"""
+----------------------+--------------------------+-----------------+---------+---------+---------+---------+---------+-------+
|         rpc          |           name           |       user      |   mean  |   50%   |   90%   |   99%   |   max   | iters |
+----------------------+--------------------------+-----------------+---------+---------+---------+---------+---------+-------+
|   authorize_query    |      speed of light      |      admin      |  0.363  |  0.278  |  0.338  |  8.051  |  8.051  |  100  |
| get_protocol_version |   get_protocol_version   |      admin      |  0.134  |  0.131  |  0.141  |  0.200  |  0.200  |  100  |
|  get_server_version  |    get_server_version    |      admin      |  0.157  |  0.154  |  0.163  |  0.226  |  0.226  |  100  |
|   set_application    |     set_application      |      admin      |  0.144  |  0.142  |  0.148  |  0.245  |  0.245  |  100  |
| get_protocol_version |   get_protocol_version   | many_attrs_user |  0.142  |  0.139  |  0.146  |  0.211  |  0.211  |  100  |
|  get_server_version  |    get_server_version    | many_attrs_user |  0.159  |  0.160  |  0.168  |  0.221  |  0.221  |  100  |
|   set_application    |     set_application      | many_attrs_user |  0.127  |  0.126  |  0.129  |  0.187  |  0.187  |  100  |
|    list_datasets     | Wide table w/ attributes |      admin      |  58.085 |  57.021 |  59.195 |  94.535 |  94.535 |  100  |
|      get_schema      |   Wide table w/ attrs    |      admin      |  91.831 |  90.575 |  97.486 | 104.049 | 104.049 |  100  |
|    list_datasets     | Wide table w/ attributes |     testuser    | 157.722 | 155.523 | 158.222 | 193.572 | 193.572 |  100  |
|      get_schema      |   Wide table w/ attrs    |  pythonperfuser | 276.474 | 274.844 | 280.171 | 331.880 | 331.880 |  100  |
|     scan_as_json     | Wide table w/ many attrs |      admin      |  73.918 |  73.374 |  76.420 |  82.545 |  82.545 |  100  |
|      get_schema      | Wide table w/ many attrs |      admin      |  54.787 |  52.513 |  59.750 |  83.996 |  83.996 |  100  |
|     scan_as_json     | Wide table w/ many attrs | many_attrs_user | 499.488 | 490.157 | 512.780 | 772.903 | 772.903 |  100  |
|      get_schema      | Wide table w/ many attrs | many_attrs_user | 442.111 | 438.563 | 451.890 | 503.797 | 503.797 |  100  |
+----------------------+--------------------------+-----------------+---------+---------+---------+---------+---------+-------+

+--------------------+------------------+----------+---------+---------+---------+---------+---------+-------+
|        rpc         |       name       |   user   |   mean  |   50%   |   90%   |   99%   |   max   | iters |
+--------------------+------------------+----------+---------+---------+---------+---------+---------+-------+
| list_dataset_names | Many wide tables |  admin   |  15.462 |  14.772 |  16.967 |  24.237 |  24.237 |   50  |
| list_dataset_names | Many wide tables | testuser | 259.072 | 250.976 | 283.031 | 481.884 | 481.884 |   50  |
+--------------------+------------------+----------+---------+---------+---------+---------+---------+-------+
"""
import unittest
import pytest

from okera import policy_sync
from okera import _thrift_api
from okera._thrift_api import (
    TAccessPermissionLevel,
    TAccessPermissionScope,
    TAttributeMatchLevel,
    TAuthorizeQueryClient,
    TGetDatasetsParams,
    TPlanRequestParams,
    TRequestType,
)
from okera.tests import create_test_data
from okera.tests import pycerebro_test_common as common

from sf_test import get_snowflake_test_context, get_snowflake

ATTR = "perf_test.attr1"
DB = "python_perf_test_db"

TBL1 = "tbl1"
TBL2 = "tbl2"
ROLE = "python_perf_test_role"
TEST_USER = "pythonperfuser"

MANY_ATTRS_DB = "many_attrs_db"
MANY_ATTRS_TABLE = "many_attrs_tbl"
MANY_ATTRS_ROLE = "many_attrs_role"
MANY_ATTRS_USER = "many_attrs_user"

PERF = common.PerfResults()

def _create_wide_view(db, tbl, multiple, source_db, source_tbl):
    cols = []
    for idx in range(multiple):
        cols.extend([
            "uid AS uid%04d" % (idx),
            "dob AS dob%04d" % (idx),
            "gender AS gender%04d" % (idx),
            "ccn AS ccn%04d" % (idx),
        ])
    stmt = "CREATE VIEW %s.%s AS SELECT %s FROM %s.%s" % (
        db, tbl, ', '.join(cols), source_db, source_tbl)

    return stmt

@pytest.mark.perf
class PerfTest(common.TestBase):
    @classmethod
    def tearDownClass(cls):
        PERF.finalize_results()

    @classmethod
    def setUpClass(cls):
        if not PERF.run_perf():
            return
        if PERF.skip_data_load():
            print("Skipping data load.")
            return

        print("\n\nSetting up perf test.")

        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            print("...Creating wide view")
            ddls = [
                "DROP ATTRIBUTE IF EXISTS %s" % (ATTR),
                "CREATE ATTRIBUTE %s" % (ATTR),

                "DROP DATABASE IF EXISTS %s CASCADE" % (DB),
                "CREATE DATABASE %s" % (DB),

                """CREATE TABLE {db}.users (
                  uid STRING ATTRIBUTE {attr},
                  dob STRING ATTRIBUTE {attr},
                  gender STRING ATTRIBUTE {attr},
                  ccn STRING ATTRIBUTE {attr}
                )""".format(db=DB, attr=ATTR),

                _create_wide_view(DB, TBL1, 100, DB, 'users'),

                "DROP ROLE IF EXISTS %s" % (ROLE),
                "CREATE ROLE %s WITH GROUPS %s" % (ROLE, TEST_USER),
                "GRANT SELECT ON DATABASE %s HAVING ATTRIBUTE IN (%s) TO ROLE %s" % \
                    (DB, ATTR, ROLE),
            ]

            for ddl in ddls:
                conn.execute_ddl(ddl)

            print("...Creating xilinx test case")
            create_test_data.create_xilinx_tbl1(
                conn, MANY_ATTRS_DB, MANY_ATTRS_TABLE, MANY_ATTRS_ROLE, MANY_ATTRS_USER)
        print("...Done setting up perf test")

    @unittest.skipIf(not PERF.run_perf(), "Skipping perf tests")
    def test_list_roles(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            print("...Setting up list_roles test")
            total_roles = 150
            role_prefix = "list_roles_test"
            user_prefix = "list_roles_user"
            for i in range(total_roles):
                role_name = role_prefix + str(i)
                user_name = user_prefix + str(i)
                conn.execute_ddl("CREATE ROLE IF NOT EXISTS %s WITH GROUPS %s" % \
                                 (role_name, user_name))
                conn.execute_ddl("""
                GRANT IF NOT EXISTS SELECT ON DATABASE %s
                HAVING ATTRIBUTE IN (%s)
                TO ROLE %s
                POLICYPROPERTIES(
                    'enabled'='true',
                    'start_datetime'='543453453',
                    'end_datetime'='543453459')
                """ % (DB, ATTR, role_name))
                conn.execute_ddl("""
                GRANT IF NOT EXISTS SELECT ON TABLE %s.%s
                TO ROLE %s
                POLICYPROPERTIES('enabled'='true')
                """ % (DB, TBL1, role_name))

            def list(t_params, expected_results):
                def get():
                    t_params.return_structured_exprs = True
                    t_params.limit = 10000
                    roles = conn._underlying_client().ListRoles(t_params)
                    assert len(roles.roles) == expected_results
                return get

            print("...starting perf run for list_roles")

            t_params = _thrift_api.TListRolesParams()
            PERF.measure(list(t_params, total_roles + 7), "Listing Roles",
                         "admin", "ListRoles")

            t_params = _thrift_api.TListRolesParams()
            t_params.scopes = [TAccessPermissionScope.DATABASE]
            PERF.measure(list(t_params, total_roles + 3), "Listing Roles w/scope filter",
                         "admin", "ListRoles")

            t_params = _thrift_api.TListRolesParams()
            t_params.group_names = [
                "list_roles_user1",
                "list_roles_user125",
                "list_roles_user10",
                "no_match"]
            PERF.measure(list(t_params, 3), "Listing Roles w/group_names filter",
                         "admin", "ListRoles")

            t_params = _thrift_api.TListRolesParams()
            t_params.role_name_substring = role_prefix
            PERF.measure(list(t_params, total_roles),
                         "Listing Roles w/role name substring filter",
                         "admin", "ListRoles")

    def _init_snowflake_setup(self, conn, ddls_params):
        setup_ddls = [
            "DROP DATABASE IF EXISTS {db} CASCADE INCLUDING PERMISSIONS",
            "CREATE DATABASE {db}",
            """
              CREATE DATACONNECTION IF NOT EXISTS {conn} CXNPROPERTIES
                (
                  'connection_type' = 'JDBC',
                  'driver' = 'snowflake',
                  'account' = 'vq85960',
                  'host' = 'vq85960.snowflakecomputing.com',
                  'port'= '443',
                  'user'= 'awssm://okera/demo/snowflake/username',
                  'password' = 'awssm://okera/demo/snowflake/password',
                  'jdbc.db.name' = 'OKERA_DEMO'
                )""",
            """
              CREATE EXTERNAL TABLE {db}.hospital_discharge
              STORED AS JDBC
              TBLPROPERTIES(
                'driver' = 'snowflake',
                'okera.connection.name' = '{conn}',
                'jdbc.db.name' = 'OKERA_DEMO',
                'jdbc.schema.name' = 'HEALTHCARE',
                'table' = 'HOSPITAL_DISCHARGE'
              )
            """,
            """
              CREATE EXTERNAL TABLE {db}.patient
              STORED AS JDBC
              TBLPROPERTIES(
                'driver' = 'snowflake',
                'okera.connection.name' = '{conn}',
                'jdbc.db.name' = 'OKERA_DEMO',
                'jdbc.schema.name' = 'HEALTHCARE',
                'table' = 'PATIENT'
              )
            """,
            "DROP ROLE IF EXISTS {role}",
            "CREATE ROLE {role}",
            "GRANT ROLE {role} TO GROUP {group}",
            "GRANT SELECT ON DATABASE {db} TO ROLE {role}",
        ]

        for ddl_tmpl in setup_ddls:
            ddl = ddl_tmpl.format(**ddls_params)
            conn.execute_ddl(ddl)

    @unittest.skipIf(not PERF.run_perf_snowflake(), "Skipping snowflake perf tests")
    def test_measure_snowflake_speed_of_light(self):
        sf_ctx = get_snowflake_test_context()
        setup_ddls_params = {
            'conn' : 'snowflake_connection_demo',
            'group' : sf_ctx.username,
            'role' : 'test_perf_sf_role',
            'db' : 'test_perf_sf',
        }

        planner_ctx = common.get_test_context()
        with common.get_planner(planner_ctx) as conn:
            self._init_snowflake_setup(conn, setup_ddls_params)

        def conn_query(conn):
            def get():
                res = conn.query_request(
                    "SELECT * FROM OKERA_DEMO.HEALTHCARE.PATIENT LIMIT 1")
                assert len(res.get('rowset')) == 1
            return get

        with get_snowflake(sf_ctx) as conn:
            conn.query_request("USE ROLE {}".format(sf_ctx.user_role))
            PERF.measure(conn_query(conn), "sfproxy: speed of light",
                         sf_ctx.username, "sfproxy_query_request")

    @unittest.skipIf(not PERF.run_perf_snowflake(), "Skipping snowflake perf tests")
    def test_measure_snowflake_generate_policy_sync(self):
        sf_ctx = get_snowflake_test_context(username='SFV3ADMIN')
        setup_ddls_params = {
            'conn' : 'snowflake_connection_demo',
            'group' : sf_ctx.username,
            'role' : 'test_perf_sf_role',
            'db' : 'test_perf_sf',
        }

        planner_ctx = common.get_test_context()
        with common.get_planner(planner_ctx) as conn:
            self._init_snowflake_setup(conn, setup_ddls_params)
            request = TGetDatasetsParams()
            request.dataset_names = ['{}.{}'.format('test_perf_sf', 'patient')]
            request.with_schema = True
            okera_tbl = conn.service.client.GetDatasets(request).datasets[0]

        with get_snowflake(sf_ctx) as conn:
            users = policy_sync.get_snowflake_users(conn, before_role='ACCOUNTADMIN')

        def generate(conn):
            def get():
                res = policy_sync.generate_policies(
                    conn, users, 'test_perf_sf', 'patient', okera_tbl)
                assert res is not None
            return get

        with common.get_planner(planner_ctx) as conn:
            PERF.measure(generate(conn), "generate policy sync (all sf users)",
                         'admin', "generate_policies")

    @unittest.skipIf(not PERF.run_perf(), "Skipping perf tests")
    def test_wide_table_with_attributes(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            def list(tbl):
                def get():
                    datasets = conn.list_datasets(DB, name=tbl)
                    assert len(datasets) >= 1
                return get

            def describe(user, plan):
                def fn():
                    params = TPlanRequestParams()
                    params.request_type = TRequestType.Sql
                    params.requesting_user = user
                    params.sql_stmt = \
                        'select * from %s.%s' % (DB, TBL1)
                    if plan:
                        conn.service.client.PlanRequest(params)
                    else:
                        conn.service.client.GetSchema(params)
                return fn

            PERF.measure(list(TBL1), "Wide table w/ attributes",
                         "admin", "list_datasets")
            PERF.measure(describe(None, False), "Wide table w/ attrs",
                         "admin", "get_schema")
            PERF.measure(describe(None, True), "Wide table w/ attrs",
                         "admin", "plan_request")

            ctx.enable_token_auth(token_str=TEST_USER)
            PERF.measure(list(TBL1), "Wide table w/ attributes",
                         TEST_USER, "list_datasets")
            PERF.measure(describe(TEST_USER, False), "Wide table w/ attrs",
                         TEST_USER, "get_schema")
            PERF.measure(describe(TEST_USER, True), "Wide table w/ attrs",
                         TEST_USER, "plan_request")

    @unittest.skipIf(not PERF.run_perf(), "Skipping perf tests")
    def test_xilinx_table(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            def scan():
                conn.scan_as_json(
                    'select * from %s.%s' % (MANY_ATTRS_DB, MANY_ATTRS_TABLE))

            def describe(user, plan):
                def fn():
                    params = TPlanRequestParams()
                    params.request_type = TRequestType.Sql
                    params.requesting_user = user
                    params.sql_stmt = \
                        'select * from %s.%s' % (MANY_ATTRS_DB, MANY_ATTRS_TABLE)
                    if plan:
                        conn.service.client.PlanRequest(params)
                    else:
                        conn.service.client.GetSchema(params)
                return fn

            PERF.measure(scan, "Wide table w/ many attrs",
                         "admin", "scan_as_json")
            PERF.measure(describe(None, False), "Wide table w/ many attrs",
                         "admin", "get_schema")
            PERF.measure(describe(None, True), "Wide table w/ many attrs",
                         "admin", "plan_request")

            ctx.enable_token_auth(token_str=MANY_ATTRS_USER)
            PERF.measure(scan, "Wide table w/ many attrs",
                         MANY_ATTRS_USER, "scan_as_json")
            PERF.measure(describe(MANY_ATTRS_USER, False), "Wide table w/ many attrs",
                         MANY_ATTRS_USER, "get_schema")
            PERF.measure(describe(MANY_ATTRS_USER, True), "Wide table w/ many attrs",
                         MANY_ATTRS_USER, "plan_request")

    @unittest.skipIf(not PERF.run_perf(), "Skipping perf tests")
    def test_measure_speed_of_light(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            def get():
                self.authorize_query(
                    conn, "select 1", client=TAuthorizeQueryClient.TEST_ACK)
            PERF.measure(get, "speed of light",
                         "admin", "authorize_query")

    @unittest.skipIf(not PERF.run_perf(), "Skipping perf tests")
    def test_measure_status_apis(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            def set_application():
                conn.service.client.SetApplication('perf-test')
            def get_metric():
                conn.service.client.GetMetric('perf-test')

            PERF.measure(lambda: conn.get_protocol_version(),
                         'get_protocol_version', 'admin', 'get_protocol_version')
            PERF.measure(lambda: conn.get_server_version(),
                         'get_server_version', 'admin', 'get_server_version')
            PERF.measure(lambda: conn.set_application('perf-test'),
                         'set_application', 'admin', 'set_application')
            PERF.measure(get_metric, 'get_metric', 'admin', 'get_metric')
            ctx.enable_token_auth(token_str=MANY_ATTRS_USER)
            PERF.measure(lambda: conn.get_protocol_version(),
                         'get_protocol_version', MANY_ATTRS_USER, 'get_protocol_version')
            PERF.measure(lambda: conn.get_server_version(),
                         'get_server_version', MANY_ATTRS_USER, 'get_server_version')
            PERF.measure(lambda: conn.set_application('perf-test'),
                         'init_application', MANY_ATTRS_USER, 'init_application')
            PERF.measure(set_application,
                         'set_application', MANY_ATTRS_USER, 'set_application')

    @unittest.skipIf(not PERF.run_perf(), "Skipping perf tests")
    def test_measure_list_datbases_api(self):
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            def list_db(db=None, pattern=None, tags=None, match_level=None, privilege=None):
                def get():
                    if db is not None:
                        result = conn.list_databases_v2(exact_names_filter=[db])
                        assert len(result.databases) >= 1
                    if pattern is not None:
                        result = conn.list_databases_v2(name_pattern_filter=pattern)
                        assert len(result.databases) >= 1
                    if privilege is not None:
                        result = conn.list_databases_v2(privilege=privilege, limit=1000)
                        returned_dbs = []
                        for database in result.databases:
                            returned_dbs.append(database.name)
                            print("shashi : %s" % database.name)
                        assert ([DB] in returned_dbs)
                    if tags is not None:
                        result = conn.list_databases_v2(
                            tags=tags, tag_match_level=match_level)
                        assert len(result.databases) >= 1
                return get

            pattern = '*perf_test*'
            attr = common.TestBase._get_t_attribute_obj('perf_test', 'attr1')
            priv = TAccessPermissionLevel.SELECT

            PERF.measure(list_db(db=DB), "Exact name filter",
                         "admin", "list_databases_v2")
            PERF.measure(list_db(pattern=pattern), "Pattern filter",
                         "admin", "list_databases_v2")
            PERF.measure(list_db(privilege=priv), "Privilege filter",
                         "admin", "list_databases_v2")
            PERF.measure(list_db(tags=[attr],
                         match_level=TAttributeMatchLevel.DATABASE_PLUS),
                         "Attributes filter", "admin", "list_databases_v2")

            ctx.enable_token_auth(token_str=TEST_USER)
            PERF.measure(list_db(db=DB), "Exact name filter",
                         TEST_USER, "list_databases_v2")
            PERF.measure(list_db(pattern=pattern), "Pattern filter",
                         TEST_USER, "list_databases_v2")
            PERF.measure(list_db(privilege=priv), "Privilege filter",
                         TEST_USER, "list_databases_v2")
            PERF.measure(list_db(tags=[attr],
                         match_level=TAttributeMatchLevel.DATABASE_PLUS),
                         "Attributes filter", TEST_USER, "list_databases_v2")

    @unittest.skipIf(not PERF.run_perf(), "Skipping perf tests")
    def test_measure_list_datasets(self):
        TEST_ROLE1 = "test_list_datasets_role1"
        TEST_USER1 = "test_list_datasets_user1"
        TEST_ROLE2 = "test_list_datasets_role2"
        TEST_USER2 = "test_list_datasets_user2"
        TEST_ROLE3 = "test_list_datasets_role3"
        TEST_USER3 = "test_list_datasets_user3"
        db = 'perf_test_db_many_wide_tables'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            db = create_test_data.require(
                PERF, conn, create_test_data.DataHarness.LARGE_DB_WIDE_TABLES, db=db)
            def list(user=None):
                def get():
                    datasets = conn.list_datasets(db, requesting_user=user)
                    assert len(datasets) == 25
                return get

            self._recreate_test_role(conn, TEST_ROLE1, [TEST_USER1])
            self._recreate_test_role(conn, TEST_ROLE2, [TEST_USER2])
            self._recreate_test_role(conn, TEST_ROLE3, [TEST_USER3])

            conn.execute_ddl(('GRANT SELECT ON DATABASE %s ' +
                              'TRANSFORM %s.email WITH sha2() ' +
                              'TO ROLE %s') % (db, db, TEST_ROLE1))
            conn.execute_ddl(('GRANT SELECT ON CATALOG ' +
                              'TRANSFORM %s.email WITH sha2() ' +
                              'TO ROLE %s') % (db, TEST_ROLE2))
            conn.execute_ddl(('GRANT SELECT ON CATALOG HAVING ATTRIBUTE %s.email ' +
                              'TRANSFORM %s.email WITH sha2() ' +
                              'TO ROLE %s') % (db, db, TEST_ROLE3))

            PERF.measure(list(), "multiple wide tables in db",
                         "admin", "list_datasets")
            PERF.measure(list(TEST_USER1), "multiple wide tables in db",
                         TEST_USER1, "list_datasets")
            PERF.measure(list(TEST_USER2), "multiple wide tables in db",
                         TEST_USER2, "list_datasets")
            #PERF.measure(list(TEST_USER3), "multiple wide tables in db",
            #             TEST_USER3, "list_datasets")

    @unittest.skipIf(not PERF.run_perf(), "Skipping perf tests")
    def test_measure_recover_partitions(self):
        db = 'perf_test_trial_analytics'
        ctx = common.get_test_context()
        with common.get_planner(ctx) as conn:
            def recover():
                self._recreate_test_db(conn, db)
                conn.execute_ddl('''
                    CREATE EXTERNAL TABLE %s.audit_logs(
                      request_time STRING
                    )
                    PARTITIONED BY (tenant string, ymd string)
                    STORED AS JSON
                    LOCATION 's3://okera-trial-us-west-2-e71015d/tenants/logs/audit/'
                    ''' % db)
                conn.execute_ddl('''
                    CREATE EXTERNAL TABLE %s.web_analytics(
                      event_time STRING
                    )
                    PARTITIONED BY (tenant string, ymd string)
                    STORED AS JSON
                    LOCATION 's3://okera-trial-us-west-2-e71015d/tenants/logs/analytics/'
                    ''' % db)
                conn.execute_ddl(
                    'ALTER TABLE %s.audit_logs RECOVER PARTITIONS' %  db)
                conn.execute_ddl(
                    'ALTER TABLE %s.web_analytics RECOVER PARTITIONS' % db)
            PERF.measure(recover, 'recover partition with many folders and files',
                         None, '', iters=5)

            def show():
                res = conn.execute_ddl('SHOW PARTITIONS %s.audit_logs' % db)
                self.assertTrue(len(res) > 2000)
            def show_files():
                res = conn.execute_ddl('SHOW FILES IN %s.audit_logs' % db)
            PERF.measure(show, 'show thousands partitions',
                         None, '')
            PERF.measure(show_files, 'show thousands files in partitions',
                         None, '')

    @unittest.skipIf(not PERF.run_perf(), "Skipping perf tests")
    def test_list_table_names(self):
        db = 'perf_test_large_db'
        ctx = common.get_test_context()
        role = 'perf_test_large_db_role'
        user = 'perf_test_large_db_user'
        with common.get_planner(ctx) as conn:
            self._recreate_test_role(conn, role, [user])
            conn.delete_attribute(db, 'financial')
            conn.create_attribute(db, 'financial')
            conn.delete_attribute(db, 'phone_number')
            conn.create_attribute(db, 'phone_number')
            conn.delete_attribute(db, 'credit_card')
            conn.create_attribute(db, 'credit_card')
            create_test_data.create_large_db_wide_tables_attributes(conn, db, 1000)
            conn.execute_ddl(('GRANT SELECT ON CATALOG ' +
                              'HAVING ATTRIBUTE NOT IN (%s.financial) ' +
                              'TRANSFORM %s.email WITH sha2() ' +
                              'TRANSFORM %s.phone_number WITH sha2() ' +
                              'TRANSFORM %s.credit_card WITH mask() ' +
                              'TRANSFORM %s.person WITH mask() ' +
                              'TO ROLE %s') % (db, db, db, db, db, role))

            def ls():
                conn.list_dataset_names(db)

            PERF.measure(ls, "Many wide tables", "admin", "list_dataset_names")

            ctx.enable_token_auth(token_str=user)
            PERF.measure(ls, "Many wide tables", "testuser", "list_dataset_names")


    # TODO: missing RPCs
    # TExecDDLResult ExecuteDDL2(1:TExecDDLParams ddl)
    # throws(1:RecordService.TRecordServiceException ex);
    # TGetInfoResult GetInfo(1: TGetInfoParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    #
    # TGetAuthenticatedUserResult AuthenticatedUser(1:string token)
    # throws(1:RecordService.TRecordServiceException ex);
    #
    # TGetDatabasesResult GetDatabases(1:TGetDatabasesParams params)
    #     throws(1:TRecordServiceException ex);
    # TGetTablesResult GetTables(1:TGetTablesParams params)
    # throws(1:TRecordServiceException ex);
    # TGetPartitionsResult GetPartitions(1:TGetPartitionsParams params)
    # throws(1:TRecordServiceException ex);
    # TGetDatasetsResult GetDatasets(1:TGetDatasetsParams params)
    # throws(1:RecordService.TRecordServiceException ex);
    #
    # TGetAccessPermissionsResult GetAccessPermissions(1: TGetAccessPermissionsParams p)
    # throws(1:RecordService.TRecordServiceException ex);
    # TGetRoleProvenanceResult GetRoleProvenance(1: TGetRoleProvenanceParams params)
    # throws(1:RecordService.TRecordServiceException ex);
    # TGetGrantableRolesResult GetGrantableRoles(1: TGetGrantableRolesParams params)
    # throws(1:RecordService.TRecordServiceException ex);
    #
    # TGetUdfsResult GetUdfs(1:TGetUdfsParams params)
    # throws(1:RecordService.TRecordServiceException ex);
    #
    # TAddRemovePartitionsResult AddRemovePartitions(1:TAddRemovePartitionsParams params)
    # throws(1:RecordService.TRecordServiceException ex);
    #
    # TListFilesResult ListFiles(1:TListFilesParams params)
    # throws(1:RecordService.TRecordServiceException ex);
    # TGetRegisteredObjectsResult GetRegisteredObjects(1: TGetRegisteredObjectsParams p)
    #  throws(1:RecordService.TRecordServiceException ex);
    #
    # TCreateAttributesResult CreateAttributes(1:TCreateAttributesParams params)
    #  throws(1:RecordService.TRecordServiceException ex);
    # TGetAttributesResult GetAttributes(1: TGetAttributesParams params)
    #  throws(1:RecordService.TRecordServiceException ex);
    # TGetCountResult GetRecordCount(1: TGetCountParams params)
    #  throws(1:RecordService.TRecordServiceException ex);
    # TAssignAttributesResult AssignAttributes(1: TAssignAttributesParams params)
    #  throws(1:RecordService.TRecordServiceException ex);
    # TUnassignAttributesResult UnassignAttributes(1: TUnassignAttributesParams params)
    #  throws(1:RecordService.TRecordServiceException ex);
    # TDeleteAttributesResult DeleteAttributes(1: TDeleteAttributesParams params)
    #  throws(1:RecordService.TRecordServiceException ex);
    # TGetAttributeNamespacesResult GetAttributeNamespaces(
    #    1: TGetAttributeNamespacesParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    # TSetAttributesResult SetAttributes(1: TSetAttributesParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    # TUpdateAttributeMappingsResult UpdateAttributeMappings(
    #    1: TUpdateAttributeMappingsParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    #
    # TConfigChangeResult UpsertConfig(1: TConfigUpsertParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    # TConfigChangeResult DeleteConfig(1: TConfigDeleteParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    #
    # TLogInfoResult LogInfo(1: TLogInfoParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    #
    # TRefreshCredentialsResult RefreshCredentials(1: TRefreshCredentialsParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    #
    # TDataRegConnectionResult ManageDataRegConnection(1: TDataRegConnectionParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    # TDiscoverCrawlerResult DiscoverCrawler(1: TDiscoverCrawlerParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    # TCrawlerResult ManageCrawler(1: TCrawlerParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    # TCrawlDatasetResult ManageCrawlerDataset(1: TCrawlDatasetParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    #
    # TAuditQueryResult AuditQuery(1: TAuditQueryParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    #
    # TListCatalogsResult ListCatalogs(1: TListCatalogsParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    # TListDatabasesResult ListDatabases(1: TListDatabasesParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    #
    # TKvStoreGetResult GetFromKvStore(1: TKvStoreGetParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    # TKvStorePutResult PutToKvStore(1: TKvStorePutParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    #
    # TGetUserAttributesResult GetUserAttributes(1: TGetUserAttributesParams params)
    #    throws(1:RecordService.TRecordServiceException ex);
    #
    # TEvalResult Eval(1: TEvalParams params)
    #    throws(1:RecordService.TRecordServiceException ex);

